from typing import Any, Dict, List, Optional, Sequence, Union

from rics.collections.misc import as_list

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from ._data_structure_io import DataStructureIO


class DictIO(DataStructureIO):
    """Implementation for dicts."""

    # TODO types here all mostly wrong. Would require a lot of overloads to fix..

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, dict)

    @staticmethod
    def names(translatable: Dict[NameType, IdType]) -> List[NameType]:
        return list(translatable)

    @staticmethod
    def extract(translatable: Dict[NameType, IdType], names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        return {name: as_list(translatable[name]) for name in names}

    @staticmethod
    def insert(
        translatable: Union[Dict[NameType, IdType], Dict[NameType, Sequence[IdType]]],
        names: List[NameType],
        tmap: TranslationMap[NameType, SourceType, IdType],
        copy: bool,
    ) -> Optional[Dict[NameType, Any]]:
        from ._resolve import resolve_io

        translated = {
            key: resolve_io(value).insert(value, [key], tmap, copy=True) if key in names else value
            for key, value in translatable.items()
        }

        if copy:
            return translated

        translatable.clear()
        translatable.update(translated)  # type: ignore[arg-type]
        return None



"""Translation of IDs with flexible formatting and name matching.

For an introduction to translation, see :ref:`translation-primer` and :ref:`mapping-primer`.


Environment variables
---------------------
.. envvar:: ID_TRANSLATION_DISABLED

    Global switch. When ``true``, the :meth:`Translator.translate`-method returns immediately.
"""

import logging

from ._translator import Translator

from .__version__ import __author__  # isort:skip
from .__version__ import __copyright__  # isort:skip
from .__version__ import __title__, __description__, __version__  # isort:skip

__all__ = [
    "Translator",
    "__version__",  # Make MyPy happy
]

logging.getLogger(__name__).addHandler(logging.NullHandler())



import warnings
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Sequence, TypeVar, Union

import pandas as pd

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from ._data_structure_io import DataStructureIO
from ._sequence import SequenceIO, translate_sequence, verify_names
from .exceptions import NotInplaceTranslatableError

T = TypeVar("T", pd.DataFrame, pd.Series)


def _cast_series(series: pd.Series) -> pd.Series:
    return series.dropna().astype(int) if series.dtype == float else series


class PandasIO(DataStructureIO):
    """Implementation for Pandas data types."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, (pd.DataFrame, pd.Series, pd.Index))

    @staticmethod
    def names(translatable: T) -> Optional[List[NameType]]:
        if isinstance(translatable, pd.DataFrame):
            return list(translatable.columns)

        if isinstance(translatable, pd.MultiIndex):
            names = translatable.names
            return list(names) if any(names) else None

        return None if translatable.name is None else [translatable.name]

    @staticmethod
    def extract(translatable: T, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        if isinstance(translatable, pd.DataFrame):
            ans = defaultdict(list)
            for i, name in enumerate(translatable.columns):
                if name in names:
                    ans[name].extend(_cast_series(translatable.iloc[:, i]))
            return dict(ans)
        elif isinstance(translatable, pd.Series):
            return SequenceIO.extract(_cast_series(translatable), names)
        elif isinstance(translatable, pd.MultiIndex):
            # This will error on duplicate names, which is probably a good thing.
            return {name: list(translatable.unique(name)) for name in names}
        elif isinstance(translatable, pd.Index):
            return SequenceIO.extract(translatable, names)

        raise TypeError(f"This should not happen: {type(translatable)=}")

    @staticmethod
    def insert(
        translatable: T, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[T]:
        if isinstance(translatable, (pd.DataFrame, pd.Series)):
            if copy:
                translatable = translatable.copy()
        else:  # Index-type translatable
            if not copy:
                raise NotInplaceTranslatableError(translatable)  # TODO(issues/170): PDEP-6 check

        if isinstance(translatable, pd.DataFrame):
            for i, name in enumerate(translatable.columns):
                if name in names:
                    translatable.iloc[:, i] = _translate_series(translatable.iloc[:, i], [name], tmap)
        elif isinstance(translatable, pd.Series):
            verify_names(len(translatable), names)
            result = _translate_series(translatable, names, tmap)
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)  # TODO(issues/170): PDEP-6 check
                translatable[:] = result
        elif isinstance(translatable, pd.MultiIndex):
            df = translatable.to_frame()
            PandasIO.insert(df, names, tmap, copy=False)
            translatable = pd.MultiIndex.from_frame(df, names=translatable.names)
        elif isinstance(translatable, pd.Index):
            translatable = pd.Index(_translate_series(translatable, names, tmap), name=translatable.name)
        else:
            raise TypeError(f"This should not happen: {type(translatable)=}")

        return translatable if copy else None


def _translate_series(
    # Not MultiIndex
    series: Union[pd.Series, pd.Index],
    names: List[NameType],
    tmap: TranslationMap[NameType, SourceType, IdType],
) -> Iterable[Optional[str]]:
    verify_names(len(series), names)

    if len(names) == 1 and len(series) > 100:
        # Optimization for large series. Suboptimal if "many" values are
        # unique. Not worth the additional overhead for small series.
        magic_dict = tmap[names[0]]
        mapping = {idx: magic_dict.get(idx) for idx in series.unique()}
        return series.map(mapping)  # type: ignore[no-any-return]
    else:
        return translate_sequence(series, names, tmap)



"""Insertion and extraction of IDs and translations."""
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Sequence

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType


class DataStructureIO:
    """Insertion and extraction of IDs and translations."""

    @staticmethod
    @abstractmethod
    def handles_type(arg: Any) -> bool:
        """Return ``True`` if the implementation handles data for the type of `arg`."""

    @staticmethod
    def names(translatable: Any) -> Optional[List[NameType]]:
        """Extract names from `translatable`.

        Args:
            translatable: Data to extract names from.

        Returns:
            A list of names to translate. Returns ``None`` if names cannot be extracted.
        """
        return translatable.name if hasattr(translatable, "name") else None

    @staticmethod
    @abstractmethod
    def extract(translatable: Any, names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        """Extract IDs from `translatable`.

        Args:
            translatable: Data to extract IDs from.
            names: List of names in `translatable` to extract IDs for.

        Returns:
            A dict ``{name, ids}``.
        """

    @staticmethod
    @abstractmethod
    def insert(
        translatable: Any, names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[Any]:
        """Insert translations into `translatable`.

        Args:
            translatable: Data to translate. Modified iff ``copy=False``.
            names: Names in `translatable` to translate.
            tmap: Translations for IDs in `translatable`.
            copy: If ``True``, modify contents of the original `translatable`. Otherwise, returns a copy.

        Returns:
            A copy of `translatable` if ``copy=True``, ``None`` otherwise.

        Raises:
            NotInplaceTranslatableError: If ``copy=False`` for a type which is not translatable in-place.
        """



from __future__ import annotations

import logging
import warnings
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from time import perf_counter
from typing import Dict, Iterable, List, Mapping, Optional, Tuple, Union, final

from rics.action_level import ActionLevel, ActionLevelHelper
from rics.collections.dicts import reverse_dict
from rics.misc import tname
from rics.performance import format_seconds

from .._tasks import generate_task_id
from ..offline.types import SourcePlaceholderTranslations
from ..settings import logging as settings
from ..types import IdType, SourceType
from . import Fetcher, exceptions
from .types import IdsToFetch

LOGGER = logging.getLogger(__package__).getChild("MultiFetcher")

FetchResult = Tuple[int, SourcePlaceholderTranslations[SourceType]]


_ACTION_LEVEL_HELPER = ActionLevelHelper(
    duplicate_translation_action=ActionLevel.IGNORE,
    duplicate_source_discovered_action=None,
)


class MultiFetcher(Fetcher[SourceType, IdType]):
    """Fetcher which combines the results of other fetchers.

    Args:
        *fetchers: Fetchers to wrap.
        max_workers: Number of threads to use for fetching. Fetch instructions will be dispatched using a
             :py:class:`~concurrent.futures.ThreadPoolExecutor`. Individual fetchers will be called at most once per
             ``fetch()`` or ``fetch_all()`` call made with the ``MultiFetcher``.
        duplicate_translation_action: Action to take when multiple fetchers return translations for the same source.
        duplicate_source_discovered_action: Action to take when multiple fetchers claim the same source.
        optional_fetcher_discarded_log_level: Log level used when discarding optional fetchers for any reason.
    """

    def __init__(
        self,
        *fetchers: Fetcher[SourceType, IdType],
        max_workers: int = 1,
        duplicate_translation_action: ActionLevel.ParseType = ActionLevel.WARN,
        duplicate_source_discovered_action: ActionLevel.ParseType = ActionLevel.WARN,
        optional_fetcher_discarded_log_level: Union[int, str] = "DEBUG",
    ) -> None:
        for pos, f in enumerate(fetchers):
            if not isinstance(f, Fetcher):  # pragma: no cover
                raise TypeError(f"Argument {pos} is of type {type(f)}, expected Fetcher subtype.")

        self._id_to_rank: Dict[int, int] = {id(f): rank for rank, f in enumerate(fetchers)}
        self._id_to_fetcher: Dict[int, Fetcher[SourceType, IdType]] = {id(f): f for f in fetchers}
        self.max_workers: int = max_workers
        self._duplicate_translation_action = _ACTION_LEVEL_HELPER.verify(
            duplicate_translation_action, "duplicate_translation_action"
        )
        self._duplicate_source_discovered_action = _ACTION_LEVEL_HELPER.verify(
            duplicate_source_discovered_action, "duplicate_source_discovered_action"
        )
        if isinstance(optional_fetcher_discarded_log_level, str):
            as_int = logging.getLevelName(optional_fetcher_discarded_log_level.upper())
            if not isinstance(as_int, int):
                raise ValueError(
                    f"Bad {optional_fetcher_discarded_log_level=}. Use an integer or a valid log level name."
                )
            optional_fetcher_discarded_log_level = as_int
        self._optional_discard_level = optional_fetcher_discarded_log_level

        if len(self.fetchers) != len(fetchers):
            raise ValueError("Repeat fetcher instance(s)!")  # pragma: no cover

        self._placeholders: Optional[Dict[SourceType, List[str]]] = None
        self._source_to_id: Dict[SourceType, int] = {}

    @property
    def allow_fetch_all(self) -> bool:
        return all(f.allow_fetch_all for f in self._id_to_fetcher.values())  # pragma: no cover

    @property
    def online(self) -> bool:
        return all(f.online for f in self._id_to_fetcher.values())  # pragma: no cover

    @property
    def fetchers(self) -> List[Fetcher[SourceType, IdType]]:
        """Return child fetchers."""
        return list(self._id_to_fetcher.values())

    @final
    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        if self._placeholders is None:
            self.initialize_sources()
            return self.placeholders

        return self._placeholders

    def initialize_sources(self, task_id: int = -1, *, force: bool = False) -> None:
        if self._placeholders is None or force:
            fid_to_placeholders = self._initialize_sources(task_id)
            self._source_to_id = self._make_source_to_id(fid_to_placeholders)

            self._placeholders = {}
            for fid, placeholders in fid_to_placeholders.items():
                placeholders = {
                    source: placeholders[source] for source in placeholders if self._source_to_id[source] == fid
                }

                if placeholders:
                    self._placeholders.update(placeholders)
                else:
                    fetcher = self._id_to_fetcher[fid]
                    pretty = f"{'optional' if fetcher.optional else 'non-optional'} {self._fmt_fetcher(fetcher)}"
                    LOGGER.warning(
                        f"Discarding {pretty}: All sources found in higher-ranking fetchers.",
                        extra=dict(task_id=task_id, placeholders=placeholders),
                    )
                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]

            if not self._id_to_fetcher:
                warnings.warn("No fetchers. See log output for more information.", UserWarning, stacklevel=1)

    def _initialize_sources(self, task_id: int) -> Dict[int, Dict[SourceType, List[str]]]:
        retval: Dict[int, Dict[SourceType, List[str]]] = {}

        for fid, fetcher in list(self._id_to_fetcher.items()):
            if fetcher.optional:
                try:
                    fetcher.initialize_sources(task_id, force=True)
                    placeholders = fetcher.placeholders
                except Exception as e:  # noqa: B902
                    LOGGER.log(
                        self._optional_discard_level,
                        "Discarding optional %s: Raised\n    %s\nwhen getting sources.",
                        self._fmt_fetcher(fid),
                        f"{type(e).__name__}: {e}",
                    )
                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]
                    continue
            else:
                fetcher.initialize_sources(task_id, force=True)
                placeholders = fetcher.placeholders

                if len(placeholders) == 0:
                    level = self._optional_discard_level if fetcher.optional else logging.WARNING
                    if LOGGER.isEnabledFor(level):
                        pretty = f"{'optional' if fetcher.optional else 'non-optional'} {self._fmt_fetcher(fetcher)}"
                        LOGGER.log(level, f"Discarding {pretty}: No sources.")
                    fetcher.close()
                    del self._id_to_rank[fid]
                    del self._id_to_fetcher[fid]
                    continue

            retval[fid] = placeholders

        return retval

    def _make_source_to_id(self, fid_to_placeholders: Mapping[int, Iterable[SourceType]]) -> Dict[SourceType, int]:
        if not fid_to_placeholders:
            return {}

        source_ranks: Dict[SourceType, int] = {}
        retval: Dict[SourceType, int] = {}

        for fid, sources in fid_to_placeholders.items():
            rank = self._id_to_rank[fid]
            for source in sources:
                if source in retval:
                    self._log_rejection(source, rank, source_ranks[source], translation=False)
                else:
                    retval[source] = fid
                    source_ranks[source] = rank

        return retval

    def fetch(
        self,
        ids_to_fetch: Iterable[IdsToFetch[SourceType, IdType]],
        placeholders: Iterable[str] = (),
        *,
        required: Iterable[str] = (),
        task_id: int = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        if task_id is None:
            task_id = generate_task_id()

        tasks: Dict[int, List[IdsToFetch[SourceType, IdType]]] = {}
        sources = []
        for idt in ids_to_fetch:
            tasks.setdefault(self._source_to_id[idt.source], []).append(idt)
            sources.append(idt.source)

        placeholders = tuple(placeholders)
        required = tuple(required)

        n_sources_and_fetchers = f"{len(sources)} sources using {len(tasks)} different fetchers"

        start = perf_counter()
        log_level = settings.MULTI_FETCH
        if LOGGER.isEnabledFor(log_level.enter):
            event_key = f"{self.__class__.__name__.upper()}.FETCH"
            LOGGER.log(
                log_level.enter,
                f"Dispatch FETCH jobs for {n_sources_and_fetchers} on {self.max_workers} threads.",
                extra=dict(
                    task_id=task_id,
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    sources=sources,
                    placeholders=placeholders,
                    required_placeholders=required,
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        def fetch(fid: int) -> FetchResult[SourceType]:
            fetcher = self._id_to_fetcher[fid]
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug(f"Begin FETCH job for {len(tasks[fid])} sources using {self._fmt_fetcher(fetcher)}.")

            result = fetcher.fetch(
                tasks[fid],
                placeholders,
                required=required,
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )
            return fid, result

        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch, fid) for fid in tasks]
            ans = self._gather(futures)

        if LOGGER.isEnabledFor(log_level.exit):
            execution_time = perf_counter() - start
            LOGGER.log(
                log_level.exit,
                f"Completed FETCH jobs for {n_sources_and_fetchers} in {format_seconds(execution_time)}.",
                extra=dict(
                    task_id=task_id,
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    sources=len(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(tasks),
                    fetch_all=False,
                ),
            )

        return ans

    def fetch_all(
        self,
        placeholders: Iterable[str] = (),
        *,
        required: Iterable[str] = (),
        task_id: int = None,
        enable_uuid_heuristics: bool = False,
    ) -> SourcePlaceholderTranslations[SourceType]:
        if task_id is None:
            task_id = generate_task_id()

        placeholders = tuple(placeholders)
        required = tuple(required)

        start = perf_counter()
        log_level = settings.MULTI_FETCH_ALL
        event_key = f"{self.__class__.__name__.upper()}.FETCH_ALL"
        if LOGGER.isEnabledFor(log_level.enter):
            LOGGER.log(
                log_level.enter,
                f"Dispatch FETCH_ALL jobs for {len(self.fetchers)} fetchers on {self.max_workers} threads.",
                extra=dict(
                    event_key=event_key,
                    event_stage="ENTER",
                    event_title=f"{event_key}.ENTER",
                    placeholders=placeholders,
                    required_placeholders=required,
                    max_workers=self.max_workers,
                    num_fetchers=len(self.fetchers),
                    fetch_all=True,
                    task_id=task_id,
                ),
            )

        debug_logging_enabled = LOGGER.isEnabledFor(logging.DEBUG)

        def fetch_all(fetcher: Fetcher[SourceType, IdType]) -> FetchResult[SourceType]:
            if debug_logging_enabled:
                LOGGER.debug(f"Begin FETCH_ALL job using {self._fmt_fetcher(fetcher)}.")

            result = fetcher.fetch_all(
                placeholders,
                required=required,
                task_id=task_id,
                enable_uuid_heuristics=enable_uuid_heuristics,
            )
            return id(fetcher), result

        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix=tname(self)) as executor:
            futures = [executor.submit(fetch_all, fetcher) for fetcher in self.fetchers]
            ans = self._gather(futures)

        if LOGGER.isEnabledFor(log_level.exit):
            execution_time = perf_counter() - start
            LOGGER.log(
                log_level.exit,
                f"Completed FETCH_ALL jobs for {len(ans)} sources using "
                f"{len(self.fetchers)} fetchers in {format_seconds(execution_time)}.",
                extra=dict(
                    event_key=event_key,
                    event_stage="EXIT",
                    event_title=f"{event_key}.EXIT",
                    execution_time=execution_time,
                    sources=list(ans),
                    max_workers=self.max_workers,
                    num_fetchers=len(self.fetchers),
                    fetch_all=True,
                    task_id=task_id,
                ),
            )
        return ans

    @property
    def duplicate_translation_action(self) -> ActionLevel:
        """Return action to take when multiple fetchers return translations for the same source."""
        return self._duplicate_translation_action

    @property
    def duplicate_source_discovered_action(self) -> ActionLevel:
        """Return action to take when multiple fetchers claim the same source."""
        return self._duplicate_source_discovered_action

    def _gather(self, futures: Iterable[Future[FetchResult[SourceType]]]) -> SourcePlaceholderTranslations[SourceType]:
        ans: SourcePlaceholderTranslations[SourceType] = {}
        source_ranks: Dict[SourceType, int] = {}

        for future in as_completed(futures):
            fid, translations = future.result()
            rank = self._id_to_rank[fid]
            self._process_future_result(translations, rank, source_ranks, ans)
        return ans

    def _process_future_result(
        self,
        translations: SourcePlaceholderTranslations[SourceType],
        rank: int,
        source_ranks: Dict[SourceType, int],
        ans: SourcePlaceholderTranslations[SourceType],
    ) -> None:
        for source_translations in translations.values():
            source = source_translations.source
            other_rank = source_ranks.setdefault(source, rank)

            if other_rank != rank:
                self._log_rejection(source, rank, other_rank, translation=True)
                if rank > other_rank:
                    continue  # Don't save -- other rank is greater (lower-is-better).

            ans[source] = source_translations

    def _log_rejection(self, source: SourceType, rank0: int, rank1: int, translation: bool) -> None:  # pragma: no cover
        accepted_rank, rejected_rank = (rank0, rank1) if rank0 < rank1 else (rank1, rank0)

        rank_to_id = reverse_dict(self._id_to_rank)
        accepted = self._fmt_fetcher(rank_to_id[accepted_rank])
        rejected = self._fmt_fetcher(rank_to_id[rejected_rank])

        msg = (
            f"Discarded translations for {source=} retrieved from {rejected} since the {accepted} returned "
            "translations for the same source."
            if translation
            else f"Discarded {source=} retrieved from {rejected} since the {accepted} already claimed same source."
        )

        msg += " Hint: Rank is determined input order at initialization."

        action = self.duplicate_translation_action if translation else self.duplicate_source_discovered_action

        if action is ActionLevel.IGNORE:
            LOGGER.debug(msg)
        else:
            if action is ActionLevel.RAISE:
                LOGGER.error(msg)
                raise exceptions.DuplicateSourceError(msg)
            else:
                warnings.warn(msg, exceptions.DuplicateSourceWarning, stacklevel=2)
                LOGGER.warning(msg)

    def __repr__(self) -> str:
        max_workers = self.max_workers
        fetchers = "\n    ".join(map(str, self._id_to_fetcher.values()))
        return f"{tname(self)}({max_workers=}, fetchers=[\n    {fetchers}\n])"

    def _fmt_fetcher(self, fetcher: Union[int, Fetcher[SourceType, IdType]]) -> str:
        """Format a managed fetcher with rank and hex ID."""
        if isinstance(fetcher, int):
            fetcher = self._id_to_fetcher[fetcher]
        fetcher_id = id(fetcher)
        rank = self._id_to_rank[fetcher_id]
        return f"rank-{rank} fetcher {fetcher} at {hex(fetcher_id)}"



from typing import Any, Dict, List, Optional, Sequence, Set

from ..offline import TranslationMap
from ..types import IdType, NameType, SourceType
from ._data_structure_io import DataStructureIO


class SetIO(DataStructureIO):
    """Implementation for dicts."""

    @staticmethod
    def handles_type(arg: Any) -> bool:
        return isinstance(arg, set)

    @staticmethod
    def extract(translatable: Set[IdType], names: List[NameType]) -> Dict[NameType, Sequence[IdType]]:
        if len(names) != 1:  # pragma: no cover
            raise ValueError("Length of names must be one.")

        return {names[0]: list(translatable)}

    @staticmethod
    def insert(
        translatable: Set[IdType], names: List[NameType], tmap: TranslationMap[NameType, SourceType, IdType], copy: bool
    ) -> Optional[Set[Optional[str]]]:
        magic_dict = tmap[names[0]]
        translated = {magic_dict.get(e) for e in translatable}

        if copy:
            return translated

        translatable.clear()
        translatable.update(translated)  # type: ignore[arg-type]
        return None



"""Various supporting functions and classes."""
from ._base_metadata import BaseMetadata
from ._config_utils import ConfigMetadata
from ._load_toml import load_toml_file

__all__ = [
    "BaseMetadata",
    "ConfigMetadata",
    "load_toml_file",
]



import logging
import pickle  # noqa: S403
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional

import pandas as pd

from ..offline.types import PlaceholderTranslations, SourcePlaceholderTranslations
from ..types import ID, HasSources, IdType, SourceType
from ..utils import BaseMetadata


@dataclass(eq=False)
class CacheMetadata(BaseMetadata, Generic[SourceType, IdType], HasSources[SourceType]):
    """Metadata pertaining to fetcher caching logic.

    Args:
        cache_keys: Hierarchical identifiers for the cache. The first key is used to determine storage location, while
            the rest are used to detect configuration changes (which invalidate the cache). A typical key would be
            ``[config-file-name, config-file-sha]``.
        placeholders: A Source-to-placeholder dict.
        **kwargs: Forwarded to base classes.
    """

    def __init__(
        self,
        *,
        cache_keys: List[str],
        placeholders: Dict[SourceType, List[str]],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if not cache_keys:  # pragma: no cover
            raise ValueError("Must specify at least one cache key.")
        self._cache_keys = cache_keys
        self._placeholders = placeholders

    @property
    def cache_keys(self) -> List[str]:
        """Yields hierarchical cache keys for this metadata."""
        return self._cache_keys.copy()

    @property
    def placeholders(self) -> Dict[SourceType, List[str]]:
        return self._placeholders

    def _is_equivalent(self, other: "BaseMetadata") -> str:
        assert isinstance(other, CacheMetadata)  # noqa: S101

        if self._cache_keys != other._cache_keys:
            return f"Expected cache_keys={self._cache_keys} (requested) but got {other._cache_keys}"

        if list(self.placeholders) != list(other.placeholders):
            return f"Expected sources={list(self.placeholders)} (requested) but got {list(other.placeholders)}"

        for source, expected in self.placeholders.items():
            if sorted(expected) != sorted(other.placeholders[source]):
                return f"For {source=}, expected placeholders={expected} but got {other.placeholders[source]}"

        return ""

    def _serialize(self, to_json: Dict[str, Any]) -> Dict[str, Any]:
        ans = to_json.copy()
        to_json.clear()
        return ans

    @classmethod
    def _deserialize(cls, from_json: Dict[str, Any]) -> Dict[str, Any]:
        return dict(cache_keys=from_json.pop("_cache_keys"), placeholders=from_json.pop("_placeholders"))


class CacheAccess(Generic[SourceType, IdType]):
    """Utility class for managing the FETCH_ALL-cache.

    Args:
        max_age: Cache timeout.
        metadata: Metadata object used to determine cache validity.
    """

    CLEAR_CACHE_EXCEPTION_TYPES = (pickle.UnpicklingError,)
    """Error types which trigger cache deletion."""

    @classmethod
    def base_cache_dir_for_all_fetchers(cls) -> Path:
        """Top-level cache dir for all fetchers managed by any ``CacheAccess``-instance."""
        return Path.home().absolute().joinpath(".cache/id-translation/cached-fetcher-data/")

    @classmethod
    def clear_all_cache_data(cls) -> None:
        """Remove the entire cache directory tree for ALL instances."""
        import shutil

        cache_dir = cls.base_cache_dir_for_all_fetchers()
        print(f"Deleting the common cache directory tree at:\n    '{cache_dir}'")
        shutil.rmtree(cache_dir)

    def __init__(
        self,
        max_age: pd.Timedelta,
        metadata: CacheMetadata[SourceType, IdType],
    ) -> None:
        if max_age < pd.Timedelta(0):
            raise ValueError("fetch_all_cache_max_age must be non-negative")  # pragma: no cover
        self._max_age = max_age.to_pytimedelta()  # Avoids serializing pandas types.
        self._metadata: CacheMetadata[SourceType, IdType] = metadata
        top_level_key = next(iter(self._metadata.cache_keys))
        self._base_dir = self.base_cache_dir_for_all_fetchers().joinpath(top_level_key)
        self._logger = logging.getLogger(__package__).getChild("CacheAccess").getChild(top_level_key)

    @property
    def cache_dir(self) -> Path:
        """Get the cache directory used by this ``CacheAccess``.

        Created from :meth:`base_cache_dir_for_all_fetchers` and the first value of :attr:`CacheMetadata.cache_keys`.

        Returns:
            Cache dir for a single fetcher.
        """
        return self._base_dir

    @property
    def metadata_path(self) -> Path:
        return self._base_dir / "metadata.json"

    @property
    def data_dir(self) -> Path:
        return self._base_dir / "sources"

    def source_path(self, source: Any) -> Path:
        """Get the data path for `source`."""
        return self.data_dir / f"{source}.pkl"

    def read_cache(self, source: SourceType) -> Optional[PlaceholderTranslations[SourceType]]:
        """Read cached translation data for `source`."""
        use_cached, reason = self._metadata.use_cached(self.metadata_path, self._max_age)
        if not use_cached:
            return None

        self._logger.info(f"Use cached data for {source=}: {reason}. Cache dir: '{self.cache_dir}'.")

        path = self.source_path(source)
        try:
            with path.open("rb") as f:
                records = pickle.load(f)  # noqa: S301
        except self.CLEAR_CACHE_EXCEPTION_TYPES as e:
            self.clear(
                f"Deserializing of {source=} failed; the cache is likely corrupted."
                f"\n    -      Path: '{path}'"
                f"\n    - Exception: {type(e).__qualname__}: {e}",
                log_level=logging.ERROR,
                exc_info=True,
            )
            return None

        placeholders = tuple(self._metadata.placeholders[source])
        id_pos = placeholders.index(ID) if ID in placeholders else -1
        return PlaceholderTranslations(source, placeholders=placeholders, records=records, id_pos=id_pos)

    def write_cache(self, data: SourcePlaceholderTranslations[SourceType]) -> None:
        """Overwrite the current cache (if it exists) with new and update metadata."""
        self.data_dir.mkdir(parents=True, exist_ok=True)
        self.metadata_path.write_text(self._metadata.to_json())

        for source, pht in data.items():
            with self.source_path(source).open("wb") as f:
                pickle.dump(pht.records, f)

    def clear(self, reason: str, log_level: int = logging.DEBUG, *, exc_info: bool = False) -> None:
        """Remove cached data for the current instance."""
        deleted = [*self.cache_dir.rglob("*")]
        shutil.rmtree(self.cache_dir)

        if self._logger.isEnabledFor(log_level):
            delete_info = "\n".join(map("    - {}".format, deleted)) if deleted else "    There is nothing to delete."
            self._logger.log(
                log_level,
                f"Clear cache: {reason}.\nThe following files have been deleted:\n{delete_info}",
                exc_info=exc_info,
            )



"""Classes used for ID and translation transformation."""
import typing as _t

from ..types import IdType as _IdType


@_t.runtime_checkable
class Transformer(_t.Protocol[_IdType]):
    """Transformation API type.

    Transformers are persistent entities owned by a single :class:`.Translator` instance.

    Implementing :attr:`try_add_missing_key` is optional. Raise :class:`TransformerStop` to prevent calling the
    method multiple times.
    """

    def update_ids(self, ids: _t.Set[_IdType]) -> None:
        """Transform a source-to-ids mapping dict.

        Called just before IDs are fetched from the source.

        Args:
            ids: A collection of IDs from the `source`.
        """

    def update_translations(self, translations: _t.Dict[_IdType, str]) -> None:
        """Transform a translations.

        Called by the :class:`.MagicDict` during initialization.

        Args:
            translations: A dict of real translations ``{id: translation_string}``.
        """

    def try_add_missing_key(self, key: _IdType, /, *, translations: _t.MutableMapping[_IdType, str]) -> None:
        """Attempt to create and add a translation for an unknown ID.

        Callback function used by :class:`.MagicDict` when unknown IDs are requested. Raise :class:`.TransformerStop` to
        prevent calling this method.

        Args:
            key: An ID which is not present in the `translations`.
            translations: A mutable mapping of translations. Typically, the :class:`.MagicDict` caller itself.

        Raises:
            TransformerStop: If this method should not be called (again).
        """
        raise TransformerStop("not implemented")


class TransformerStop(ValueError):
    """Error indicating that this transformer method should not be called again.

    Transformers may raise this exception at any point, after which the method of that instance will not be called again
    by the caller which caught the exception.
    """



import logging
from typing import Any, Iterator, MutableMapping, Optional, Tuple

from rics.misc import tname

from .. import _uuid_utils
from ..transform.types import Transformer, TransformerStop
from ..types import IdType
from .types import TranslatedIds


class MagicDict(MutableMapping[IdType, str]):
    """Dictionary type for translated IDs.

    If `default_value` is given, it is used as the default answer for any calls to ``__getitem__`` where the key is
    not in `translated_ids`.

    Args:
        real_translations: A dict holding real translations.
        default_value: A string with exactly one or zero placeholders.
        enable_uuid_heuristics: Enabling may improve matching when :py:class:`~uuid.UUID`-like IDs are in use.
        transformer: Initialized :class:`.Transformer` instance.

    Examples:
        Behaviour with :attr:`default_value`.

        >>> magic = MagicDict(
        ...    {1999: "1999:Sofia", 1991: "1991:Richard"},
        ...    default_value="<Failed: id={!r}>",
        ... )
        >>> magic
        {1999: '1999:Sofia', 1991: '1991:Richard'}

        Calls to ``__getitem__`` and ``__contains__`` will never return ``False``.

        >>> magic[1999], 1999 in magic
        ('1999:Sofia', True)
        >>> magic["1999"], "1999" in magic
        ("<Failed: id='1999'>", True)

        Special handling for :py:class:`uuid.UUID` and ``UUID``-like strings improve matching.

        >>> from uuid import UUID
        >>> string_uuid = "550e8400-e29b-41d4-a716-446655440000"
        >>> magic = MagicDict(
        ...    {string_uuid: "Found!"},
        ...    enable_uuid_heuristics=True,
        ... )
        >>> magic
        {UUID('550e8400-e29b-41d4-a716-446655440000'): 'Found!'}

        >>> magic[string_uuid], magic[UUID(string_uuid)]
        ('Found!', 'Found!')

        Converting to a regular ``dict``.

        >>> dict(magic)
        {UUID('550e8400-e29b-41d4-a716-446655440000'): 'Found!'}

        Casting to ``dict`` removes all special handling.
    """

    LOGGER = logging.getLogger(__package__).getChild("MagicDict")

    def __init__(
        self,
        real_translations: TranslatedIds[IdType],
        default_value: str = None,
        enable_uuid_heuristics: bool = True,
        transformer: Transformer[IdType] = None,
    ) -> None:
        if enable_uuid_heuristics and real_translations:
            real_translations, enable_uuid_heuristics = _try_stringify_many(real_translations)

        self._real: TranslatedIds[IdType] = real_translations
        self._default = default_value
        self._cast_key = enable_uuid_heuristics

        self._try_add_missing_key = None
        if transformer is not None:
            transformer.update_translations(real_translations)
            self._try_add_missing_key = transformer.try_add_missing_key

    @property
    def default_value(self) -> Optional[str]:
        """Return the default string value to return for unknown keys, if any."""
        return self._default

    def _try_stringify(self, key: IdType) -> Any:
        return _uuid_utils.try_cast_one(key) if self._cast_key else key

    def __getitem__(self, key: IdType) -> str:
        key = self._on_read(key)
        if key in self._real or self.default_value is None:
            return self._real[key]

        return self.default_value.format(key)

    def __contains__(self, key: Any) -> bool:
        key = self._on_read(key)
        return self.default_value is not None or key in self._real

    def _on_read(self, key: IdType) -> IdType:
        key = self._try_stringify(key)
        if self._try_add_missing_key and key not in self._real:
            try:
                self._try_add_missing_key(key, translations=self)
            except TransformerStop as e:
                if self.LOGGER.isEnabledFor(logging.DEBUG):
                    call = f"{tname(self._try_add_missing_key, prefix_classname=True)}({key!r}, translations=self)"
                    self.LOGGER.debug(f"try_add_missing_key: {call} raised {e!r}. Dropping this callback function.")
                self._try_add_missing_key = None
        return key

    def __setitem__(self, key: IdType, value: str) -> None:
        key = self._on_write(key)
        self._real[key] = value

    def __delitem__(self, key: IdType) -> None:
        key = self._on_write(key)
        del self._real[key]

    def _on_write(self, key: IdType) -> IdType:
        key = self._try_stringify(key)
        return key

    def __len__(self) -> int:
        return len(self._real)

    def __iter__(self) -> Iterator[IdType]:
        return iter(self._real)

    def __repr__(self) -> str:
        return repr(self._real)


def _try_stringify_many(real_translations: TranslatedIds[IdType]) -> Tuple[TranslatedIds[IdType], bool]:
    original_ids = list(real_translations)
    try:
        new_keys = _uuid_utils.cast_many(original_ids)
    except (ValueError, AttributeError, TypeError):
        return real_translations, False  # Keys are not UUID-like.

    uuid_translations = {uuid: real_translations[idx] for uuid, idx in zip(new_keys, original_ids)}
    if len(real_translations) != len(uuid_translations):
        raise TypeError("Duplicate UUIDs found. Verify translation sources or set enable_uuid_heuristics=False.")

    return uuid_translations, True



import json
import sys
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, Tuple

from rics.performance import format_seconds


def _initialize_versions() -> Dict[str, str]:
    from pandas import __version__ as pandas_version
    from rics import __version__ as rics
    from sqlalchemy import __version__ as sqlalchemy

    from .. import __version__ as id_translation

    ans = dict(
        rics=rics,
        id_translation=id_translation,
        sqlalchemy=sqlalchemy,
        pandas=pandas_version,
    )

    if sys.version_info < (3, 11):  # pragma: no cover
        import tomli

        ans["tomli"] = tomli.__version__
    return ans


class BaseMetadata(ABC):
    """Base implementation for Metadata types.

    Args:
        versions: Top-level dependency versions.
        created: The time at which the metadata was originally created.
    """

    def __init__(self, versions: Dict[str, str] = None, created: datetime = None) -> None:
        self.versions = versions or _initialize_versions()
        self.created = created or datetime.now()

    @abstractmethod
    def _serialize(self, to_json: Dict[str, Any]) -> Dict[str, Any]:
        """Turn `to_json` into JSON-serializable types."""

    @classmethod
    @abstractmethod
    def _deserialize(cls, from_json: Dict[str, Any]) -> Dict[str, Any]:
        """Turn `from_json` into desired types."""

    @abstractmethod
    def _is_equivalent(self, other: "BaseMetadata") -> str:
        """Implementation-specific equivalence check."""

    def is_equivalent(self, other: "BaseMetadata") -> str:
        """Equivalency status."""
        if not isinstance(other, self.__class__):
            return f"Expected class={self.__class__.__name__} but got {other.__class__}"

        for package, version in self.versions.items():
            other_version = other.versions.get(package)
            if other_version != version:
                return f"Expected {package}=={version!r} (your environment) but got {package}=={other_version!r}"

        return self._is_equivalent(other)

    def to_json(self) -> str:
        """Get a JSON representation of this ``BaseMetadata``."""
        raw = self.__dict__.copy()
        kwargs = dict(
            versions=raw.pop("versions"),
            created=raw.pop("created").isoformat(),
        )
        kwargs.update(self._serialize(raw))
        assert not raw, f"Not serialized: {raw}."  # noqa:  S101
        return json.dumps(kwargs, indent=4)

    @classmethod
    def from_json(cls, s: str) -> "BaseMetadata":
        """Create ``BaseMetadata`` from a JSON string `s`."""
        raw = json.loads(s)

        kwargs = dict(
            versions=raw.pop("versions"),
            created=datetime.fromisoformat(raw.pop("created")),
        )
        kwargs.update(cls._deserialize(raw))

        assert not raw, f"Not deserialized: {raw}."  # noqa:  S101
        return cls(**kwargs)

    def use_cached(self, metadata_path: Path, max_age: timedelta) -> Tuple[bool, str]:
        """Check status of stored metadata config based a desired configuration ``self``.

        Args:
            metadata_path: Path of stored metadata.
            max_age: Maximum age of stored metadata.

        Returns:
            A tuple ``(use_cached, reason)``.
        """
        if not metadata_path.exists():
            return False, "no cache metadata found"

        stored_config = self.from_json(metadata_path.read_text())

        reason_not_equivalent = self.is_equivalent(stored_config)
        if reason_not_equivalent:
            return False, f"cached instance is not equivalent: {reason_not_equivalent}"

        expires_at = (stored_config.created + max_age).replace(microsecond=0)
        offset = format_seconds(round(abs(datetime.now() - expires_at).total_seconds()))

        if expires_at <= stored_config.created:
            return False, f"expired at {expires_at.isoformat()} ({offset} ago)"
        else:
            return True, f"expires at {expires_at.isoformat()} (in {offset})"



"""Types used for mapping."""
import typing as _t

HL = _t.TypeVar("HL", bound=_t.Hashable)
"""_t.Hashable type on the left side of a directional relationship."""
HR = _t.TypeVar("HR", bound=_t.Hashable)
"""_t.Hashable type on the right side of a directional relationship."""
LeftToRight = _t.Dict[HL, _t.Tuple[HR, ...]]
"""A left-to-right mapping."""
RightToLeft = _t.Dict[HR, _t.Tuple[HL, ...]]
"""A right-to-left mapping."""

ValueType = _t.TypeVar("ValueType", bound=_t.Hashable)
"""A type of item being mapped."""
CandidateType = _t.TypeVar("CandidateType", bound=_t.Hashable)
"""A type of item being mapped."""
MatchTuple = _t.Tuple[CandidateType, ...]
"""A tuple of candidates matched to a value."""
ContextType = _t.TypeVar("ContextType", bound=_t.Hashable)
"""Type of context in which mapping is being performed."""

UserOverrideFunction = _t.Callable[
    [ValueType, _t.Set[CandidateType], _t.Optional[ContextType]],
    _t.Optional[CandidateType],
]
"""Signature for a user-defined override function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Returns:
    Either ``None`` (let regular logic decide) or a single candidate `c` in `candidates`.
"""

ScoreFunction = _t.Callable[
    [ValueType, _t.Iterable[CandidateType], _t.Optional[ContextType]],
    _t.Iterable[float],
]
"""Signature for a likeness score function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which scoring is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Yields:
    A score for each candidate `c` in `candidates`.
"""

AliasFunction = _t.Callable[
    [ValueType, _t.Iterable[CandidateType], _t.Optional[ContextType]],
    _t.Tuple[ValueType, _t.Iterable[CandidateType]],
]
"""Signature for an alias function for heuristic scoring.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which mapping is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A tuple (name, candidates) with applied heuristics to increase (or decrease) score as desired.
"""

FilterFunction = _t.Callable[
    [ValueType, _t.Iterable[CandidateType], _t.Optional[ContextType]],
    _t.Set[CandidateType],
]
"""Signature for a filter function.

Args:
    value: An element to find matches for.
    candidates: Potential matches for `value`.
    context: The context in which filtering is being performed.

Keyword Args:
    kwargs: Accepted only by some functions.

Returns:
    A subset of candidates to keep.
"""

HeuristicsTypes = _t.Union[
    AliasFunction[ValueType, CandidateType, ContextType],
    FilterFunction[ValueType, CandidateType, ContextType],
]
"""Types that may be interpreted as a score function heuristic."""



from time import perf_counter
from typing import TYPE_CHECKING, Generic

from rics.misc import get_public_module, tname

if TYPE_CHECKING:
    from .._translator import Translator

from ..dio import resolve_io
from ..types import IdType, NameType, SourceType, Translatable


class BaseTask(Generic[NameType, SourceType, IdType]):
    """Internal type; not part of the public API."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
    ) -> None:
        self.caller = caller

        self.translatable = translatable

        self.io = resolve_io(translatable)
        self._start = perf_counter()
        self._task_id = generate_task_id(self._start)

        if caller.online:
            caller.fetcher.initialize_sources(self.task_id)

    @property
    def type_name(self) -> str:
        """Stylized typename."""
        return repr(tname(self.translatable, prefix_classname=True))

    @property
    def full_type_name(self) -> str:
        """Canonical typename."""
        clazz = type(self.translatable)
        return get_public_module(clazz) + "." + clazz.__qualname__

    @property
    def task_id(self) -> int:
        """Generated ID for this task. Used for logging."""
        return self._task_id


def generate_task_id(start: float = None) -> int:
    """Generate a new task ID."""
    return round(1000 * (perf_counter() if start is None else start))



from typing import TYPE_CHECKING, Any, Dict, Generic, List, Sequence

from rics.misc import tname

from ..transform.types import Transformer
from ..types import ID, IdType, NameType, SourceType
from ._format import Format
from ._magic_dict import MagicDict
from .types import PlaceholdersTuple, PlaceholderTranslations, TranslatedIds

if TYPE_CHECKING:
    import pandas


class FormatApplier(Generic[NameType, SourceType, IdType]):
    """Application of ``Format`` specifications.

    Args:
        translations: Matrix of ID translation components returned by fetchers.
        transformer: Initialized :class:`.Transformer` instance.

    Raises:
        ValueError: If `default` is given and any placeholder names are missing.
    """

    def __init__(
        self, translations: PlaceholderTranslations[SourceType], *, transformer: Transformer[IdType] = None
    ) -> None:
        self._translations = translations
        self._source = translations.source
        self._placeholder_names = translations.placeholders
        self._n_ids = len(translations.records)
        self._transformer = transformer

    def __call__(
        self,
        fmt: Format,
        placeholders: PlaceholdersTuple = None,
        default_fmt: Format = None,
        default_fmt_placeholders: Dict[str, Any] = None,
        enable_uuid_heuristics: bool = True,
    ) -> MagicDict[IdType]:
        """Translate IDs.

        Args:
            fmt: Translation format to use.
            placeholders: Placeholders to include in the formatted output. Use as many as possible if ``None``.
            default_fmt: Alternative format for default translation.
            default_fmt_placeholders: Default placeholders.
            enable_uuid_heuristics: Enabling may improve matching when :py:class:`~uuid.UUID`-like IDs are in use.

        Returns:
            A dict ``{idx: translated_id}``.
        """
        if placeholders is None:
            # Use as many placeholders as possible.
            placeholders = tuple(filter(self._placeholder_names.__contains__, fmt.placeholders))

        fstring = fmt.fstring(placeholders, positional=True)
        real_translations = self._apply(fstring, placeholders)

        if default_fmt is not None or default_fmt_placeholders is not None:
            if default_fmt is None:
                default_fmt = fmt
            if default_fmt_placeholders is None:
                default_fmt_placeholders = {}

            partial = default_fmt.partial(default_fmt_placeholders)
            try:
                default_fstring = partial.fstring(positional=True)
            except KeyError as e:
                raise ValueError(
                    f"All required placeholders except {{{ID}}} must be provided for {default_fmt=}:"
                    f" {default_fmt.required_placeholders}."
                ) from e
        else:
            default_fstring = None

        return MagicDict(real_translations, default_fstring, enable_uuid_heuristics, self._transformer)

    def to_dict(self) -> Dict[str, Sequence[Any]]:
        """Get the underlying data used for translations as a dict.

        Returns:
            A dict ``{placeholder: [values...]}``.
        """
        return self._translations.to_dict()

    def to_pandas(self) -> "pandas.DataFrame":
        """Get the underlying data used for translations as a :class:`pandas.DataFrame`."""
        from pandas import DataFrame

        return DataFrame(self.to_dict()).convert_dtypes()

    @property
    def records(self) -> Sequence[Sequence[Any]]:
        """Records used by this instance."""
        return self._translations.records

    def _apply(self, fstring: str, placeholders: PlaceholdersTuple) -> TranslatedIds[IdType]:
        """Apply fstring to all IDs.

        The abstract class delegates ``__apply__``-invocations to this method after some input validation.

        Args:
            fstring: A format string.
            placeholders: Keys needed for the fstring, in the order in which they appear.

        Returns:
            A dict ``{idx: translated_id}``.
        """
        id_pos, records = self._translations.id_pos, self._translations.records

        if self._placeholder_names == placeholders:
            return {record[id_pos]: fstring.format(*record) for record in records}
        else:
            pos = tuple(map(self._placeholder_names.index, placeholders))
            return {record[id_pos]: fstring.format(*(record[i] for i in pos)) for record in records}

    @property
    def source(self) -> SourceType:
        """Return translation source."""
        return self._translations.source

    @property
    def placeholders(self) -> List[str]:
        """Return placeholder names in sorted order."""
        return list(self._translations.placeholders)

    def __len__(self) -> int:
        return len(self._translations.records)

    def __repr__(self) -> str:
        placeholders = tuple(self._placeholder_names)
        source = self._source
        return f"{tname(self)}({len(self)} IDs, {placeholders=}, {source=})"



"""Types related to translation fetching."""

from dataclasses import dataclass as _dataclass
from typing import Generic as _Generic, Optional, Set

from ..offline.types import PlaceholdersTuple
from ..types import IdType, SourceType


@_dataclass(frozen=True)
class IdsToFetch(_Generic[SourceType, IdType]):
    """A source and the IDs to fetch from it."""

    source: SourceType
    """Where to fetch from."""
    ids: Set[IdType]
    """Unique IDs to fetch translations for."""


@_dataclass(frozen=True)
class FetchInstruction(_Generic[SourceType, IdType]):
    """Instructions passed from an ``AbstractFetcher`` to an implementation."""

    source: SourceType
    """Where to fetch from."""
    placeholders: PlaceholdersTuple
    """All desired placeholders in preferred order."""
    required: Set[str]
    """Placeholders that must be included in the response."""
    ids: Optional[Set[IdType]]
    """Unique IDs to fetch translations for."""
    task_id: int
    """Used for logging purposes."""
    enable_uuid_heuristics: bool
    """If set, apply heuristics to improve matching with :py:class:`~uuid.UUID`-like IDs.

    Implementations which have no UUID heuristics may silently ignore this flag.
    """

    @property
    def fetch_all(self) -> bool:
        """If ``True``, retrieve all available data."""
        return self.ids is None



"""Functions which perform heuristics for score functions.

See Also:
    The :class:`~.HeuristicScore` class.
"""
from __future__ import annotations

import logging
import re
from typing import Any as _Any, Iterable, List, Set, Tuple, Union

VERBOSE: bool = False
LOGGER = logging.getLogger(__package__).getChild("verbose").getChild("heuristic_functions")


def like_database_table(
    name: str,
    candidates: Iterable[str],
    context: _Any,
) -> Tuple[str, List[str]]:
    """Try to make `value` look like the name of a database table."""

    def apply(s: str) -> str:
        s = s.lower()
        if s == "id":
            return "id"
        s = s[: -len("id")] if s.endswith("id") else s
        s = s.replace("_", "").replace(".", "")

        if s[-1] == "s":
            pass  # Assume that any word ending in "s" is already pluralized.
        elif s[-1] in "xz" or (s[-1] == "h" and s[-2] in "sc"):
            s += "es"
        else:
            s += "s"

        return s

    return apply(name), list(map(apply, candidates))


def short_circuit(
    value: str,
    candidates: Set[str],
    context: _Any,
    value_regex: Union[str, re.Pattern[str]],
    target_candidate: str,
) -> Set[str]:
    """Short circuit `value` to the target candidate if the target and regex conditions are met.

    If `target_candidate` is in `candidates` and `value` matches the given `value_regex`, a single-element set
    ``{target_candidate}`` is returned which will trigger short-circuiting in the calling ``Mapper``. If either of
    these conditions fail, an empty set is returned and the mapping procedure will continue.

    Args:
        value: A value to map.
        candidates: Candidates for `value`.
        context: Always ignored, exists for compatibility.
        value_regex: A pattern match against `value`. Case-insensitive by default.
        target_candidate: The candidate to short circuit to.

    Returns:
        A single-element set ``{target_candidate}``, iff both conditions are met. An empty set otherwise.


    Examples:
        The main purpose of this method is to bind any name (a `value`) matching the given regex pattern to a
        pre-defined source (a `candidate`, e.g. a table in a SQL database). For example, to always match any bite
        victim-columns to the `humans` table (see the :ref:`translation-primer` page), we might specify:

        >>> value = "first_bite_victim"
        >>> value_regex= ".*_bite_victim$"  # If the column ends with '_bite_victim'..
        >>> target_candidate = "humans"  # Match it to the 'humans' table.
        >>> short_circuit(value, {"humans", "animals"}, None, value_regex, target_candidate)
        {'humans'}

        Short-circuiting will only trigger of both conditions are met. If the example below, we can see that no value is
        returned since the target candidate is not among the given candidates.

        >>> short_circuit(value, {"animals"}, None, value_regex, target_candidate)
        set()
    """
    candidates = set(candidates)
    pattern = re.compile(value_regex, flags=re.IGNORECASE) if isinstance(value_regex, str) else value_regex

    if target_candidate not in candidates:
        LOGGER.getChild("short_circuit").debug(
            f"Short-circuiting failed for {value=}: The {target_candidate=} is an input candidate."
        )
        return set()

    if not pattern.match(value):
        LOGGER.getChild("short_circuit").debug(f"Short-circuiting failed for {value=}: Does not match {pattern=}.")
        return set()

    return {target_candidate}


def force_lower_case(value: str, candidates: Iterable[str], context: _Any) -> Tuple[str, Iterable[str]]:
    """Force lower-case in `value` and `candidates`."""
    return value.lower(), list(map(str.lower, candidates))


def value_fstring_alias(
    value: str,
    candidates: Iterable[str],
    context: _Any,
    fstring: str,
    for_value: str = None,
    **kwargs: _Any,
) -> Tuple[str, Iterable[str]]:
    """Return a value formatted by `fstring`.

    Args:
        value: An element to find matches for.
        candidates: Potential matches for `value`. Not used (returned as given).
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value` and `context` as placeholders.
        for_value: If given, apply only if ``value == for_value``. When `if_value_equals` is given, `fstring` arguments
            which do not use the `value` as a placeholder key are permitted.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(formatted_value, candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'value'` and `for_value` is not given.
    """
    if not for_value and "{value}" not in fstring:
        # No longer a function of the value.
        raise ValueError(
            f"Invalid {fstring=} passed to value_fstring_alias(); does not contain {{value}}. "
            "To allow, the 'for_value' parameter must be given as well."
        )

    if for_value and value != for_value:
        return value, candidates

    return fstring.format(value=value, context=context, **kwargs), candidates


def candidate_fstring_alias(
    value: str,
    candidates: Iterable[str],
    context: _Any,
    fstring: str,
    **kwargs: _Any,
) -> Tuple[str, Iterable[str]]:
    """Return candidates formatted by `fstring`.

    Args:
        value: An element to find matches for. Not used (returned as given).
        candidates: Potential matches for `value`.
        context: Context in which the function is being called.
        fstring: The format string to use. Can use `value`, `context`, and elements of `candidates` as placeholders.
        **kwargs: Additional keyword placeholders in `fstring`.

    Returns:
        A tuple ``(value, formatted_candidates)``.

    Raises:
        ValueError: If `fstring` does not contain a placeholder `'candidate'`.
    """
    if "{candidate}" not in fstring:
        raise ValueError(f"Invalid {fstring=} passed to candidate_fstring_alias(); does not contain {{candidate}}.")

    return value, map(lambda c: fstring.format(value=value, candidate=c, context=context, **kwargs), candidates)



import logging
import warnings
from collections import defaultdict
from time import perf_counter
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Set, Union

from numpy import isnan, unique
from rics.misc import tname
from rics.performance import format_seconds

from .. import _uuid_utils
from ..dio._dict import DictIO
from ..exceptions import TooManyFailedTranslationsError
from ..mapping.types import UserOverrideFunction
from ..offline import Format, TranslationMap
from ..settings import logging as settings
from ..types import IdType, Names, NameToSource, NameType, NameTypes, SourceType, Translatable
from ..utils.logging import cast_unsafe
from ._map import MappingTask

LOGGER = logging.getLogger("id_translation.Translator.translate")

NUM_SAMPLE_IDS = 10

if TYPE_CHECKING:
    from .._translator import Translator

LOGGER = logging.getLogger("id_translation.Translator.translate")


class TranslationTask(MappingTask[NameType, SourceType, IdType]):
    """Ephemeral class for performing a single translation task on a `translatable`."""

    def __init__(
        self,
        caller: "Translator[NameType, SourceType, IdType]",
        translatable: Translatable[NameType, IdType],
        fmt: Format,
        names: Union[NameTypes[NameType], NameToSource[NameType, SourceType]] = None,
        *,
        ignore_names: Names[NameType] = None,
        override_function: UserOverrideFunction[NameType, SourceType, None] = None,
        inplace: bool = False,
        maximal_untranslated_fraction: float = 1.0,
        reverse: bool = False,
        enable_uuid_heuristics: bool = False,
    ) -> None:
        super().__init__(
            caller,
            translatable,
            names,
            ignore_names=ignore_names,
            override_function=override_function,
        )

        if not (0.0 <= maximal_untranslated_fraction <= 1):  # pragma: no cover
            raise ValueError(f"Argument {maximal_untranslated_fraction=} is not a valid fraction")

        self.fmt = fmt
        self.inplace = inplace
        self.maximal_untranslated_fraction = maximal_untranslated_fraction
        self.reverse = reverse
        self.parent: Optional[Translatable[NameType, IdType]] = None

        self.enable_uuid_heuristics = enable_uuid_heuristics

        self.key_event_level = settings.TRANSLATE_ONLINE if self.caller.online else settings.TRANSLATE_OFFLINE

    @property
    def io_names(self) -> List[NameType]:
        """Names for which IDs should be extracted from the `translatable`."""
        # Preserve input order for names, if given. These names may be repeated.
        return self.names_to_translate if self.names_from_user is None else self.names_from_user

    def extract_ids(self) -> Dict[SourceType, Set[IdType]]:
        """Extract IDs to fetch from the translatable."""
        name_to_source = self.name_to_source
        source_to_ids: Dict[SourceType, Set[IdType]] = defaultdict(set)

        float_names: List[NameType] = []
        num_coerced = 0
        ids: Sequence[IdType]
        for name, ids in self.io.extract(self.translatable, self.io_names).items():
            if len(ids) == 0:
                continue

            if isinstance(ids[0], float):
                float_names.append(name)
                # Float IDs aren't officially supported, but is common when using Pandas since int types cannot be NaN.
                # This is sometimes a problem for the built-in set (see https://github.com/numpy/numpy/issues/9358), and
                # for several database drivers.
                arr = unique(ids)  # type: ignore[var-annotated]
                keep_mask = ~isnan(arr)
                num_coerced += keep_mask.sum()  # Somewhat inaccurate; includes repeat IDs from other names
                source_to_ids[name_to_source[name]].update(arr[keep_mask].astype(int, copy=False))
            else:
                if self.enable_uuid_heuristics:
                    ids = _uuid_utils.try_cast_many(ids)

                source_to_ids[name_to_source[name]].update(ids)

        if num_coerced > 100:  # pragma: no cover
            warnings.warn(
                f"To ensure proper fetcher operation, {num_coerced} float-type IDs have been coerced to integers. "
                f"Enforcing supported data types for IDs (str and int) in your {self.type_name!r}-data may improve "
                f"performance. Affected names ({len(float_names)}): {float_names}.",
                stacklevel=3,
            )
        return source_to_ids

    def log_key_event_enter(self) -> None:
        """Emits the ``TRANSLATOR:TRANSLATE.ENTER`` message."""
        if not LOGGER.isEnabledFor(self.key_event_level.enter):
            return

        type_name = self.type_name
        names = self.names_from_user
        ignore_names = self.ignore_names

        name_info = f"Derive based on type={type_name}" if names is None else repr(names)
        if ignore_names is not None:
            name_info += f", excluding those given by {ignore_names=}"

        LOGGER.log(
            level=self.key_event_level.enter,
            msg=f"Begin translation of {type_name}-type data. Names to translate: {name_info}.",
            extra=dict(
                task_id=self.task_id,
                event_key="TRANSLATOR.TRANSLATE",
                event_stage="ENTER",
                event_title="TRANSLATOR.TRANSLATE.ENTER",
                sources=self.caller.sources,
                online=self.caller.online,
                # Task-specific
                translatable_type=self.full_type_name,
                names=names,
                ignore_names=tname(ignore_names, prefix_classname=True) if callable(ignore_names) else ignore_names,
                inplace=self.inplace,
                reverse=self.reverse,
            ),
        )

    def log_key_event_exit(self) -> None:
        """Emits the ``TRANSLATOR:TRANSLATE.EXIT`` key event message."""
        if not LOGGER.isEnabledFor(self.key_event_level.exit):
            return

        inplace = self.inplace
        n2s = self.name_to_source
        execution_time = perf_counter() - self._start

        LOGGER.log(
            level=self.key_event_level.exit,
            msg=(
                f"Finished translation of {len(n2s)} names in {self.type_name}-type data in "
                f"{format_seconds(execution_time)}, using name-to-source mapping: {n2s}."
            ),
            extra=dict(
                task_id=self.task_id,
                event_key="TRANSLATOR.TRANSLATE",
                event_stage="EXIT",
                event_title="TRANSLATOR.TRANSLATE.EXIT",
                execution_time=execution_time,
                # Task-specific
                translatable_type=self.full_type_name,
                inplace=inplace,
                reverse=self.reverse,
                name_to_source=n2s,
            ),
        )

    def insert(
        self, translation_map: TranslationMap[NameType, SourceType, IdType]
    ) -> Optional[Translatable[NameType, str]]:
        """Insert translated IDs into the `translatable`, based on data retrieved by the fetcher."""
        inplace = self.inplace

        translation_map.reverse_mode = self.reverse
        try:
            result = self.io.insert(
                self.translatable,
                names=self.io_names,
                tmap=translation_map,
                copy=not inplace,
            )
        finally:
            translation_map.reverse_mode = False

        return None if inplace else result

    def verify(self, translation_map: TranslationMap[NameType, SourceType, IdType]) -> None:
        """Verify translations.

        Performs translation with pre-defined formats, counting the number of IDs which are (and aren't) known to the
        translation map.
        """
        if not (LOGGER.isEnabledFor(logging.DEBUG) or self.maximal_untranslated_fraction < 1.0):
            pass

        name_to_mask: Dict[NameType, Sequence[Any]] = self.io.extract(self.translatable, self.io_names)

        if self.enable_uuid_heuristics:
            for name in name_to_mask:
                name_to_mask[name] = _uuid_utils.try_cast_many(name_to_mask[name])

        fmt, default_fmt = translation_map.fmt, translation_map.default_fmt
        try:
            translation_map.fmt = ""  # type: ignore[assignment]
            translation_map.default_fmt = None
            DictIO.insert(
                name_to_mask,
                names=self.io_names,
                tmap=translation_map,
                copy=False,
            )
        finally:
            translation_map.fmt, translation_map.default_fmt = fmt, default_fmt

        name_to_ids: Optional[Dict[NameType, Sequence[IdType]]] = None

        def get_ids(n: NameType) -> Sequence[Any]:
            nonlocal name_to_ids
            if name_to_ids is None:
                # It would be nice to pass just names=[n], but semantics may change depending
                # on the number of names given. Is there a way to do this safely?
                name_to_ids = self.io.extract(self.translatable, names=self.io_names)
                if self.enable_uuid_heuristics:
                    for n in name_to_ids:
                        name_to_ids[n] = _uuid_utils.try_cast_many(name_to_ids[n])
            return name_to_ids[n]

        for name, mask in name_to_mask.items():
            n_untranslated, n_total = sum(t is None for t in mask), len(mask)
            if n_untranslated == 0:
                continue
            f_untranslated = n_untranslated / n_total

            sample_ids = self._get_untranslated_ids(get_ids(name), mask=mask)

            extra = {
                "name_of_ids": name,
                "source": translation_map.name_to_source[name],
                "sample_ids": cast_unsafe(sample_ids),
            }

            message = (
                f"Failed to translate {n_untranslated}/{n_total} ({f_untranslated:.1%}{{reason}}) of IDs "
                f"for {name=} using source={translation_map.name_to_source[name]!r}. Sample IDs: {sample_ids}."
            )

            if f_untranslated > self.maximal_untranslated_fraction:
                message = message.format(
                    reason=f" > maximal_untranslated_fraction={self.maximal_untranslated_fraction:.1%}"
                )
                LOGGER.error(message, extra=extra)
                raise TooManyFailedTranslationsError(message)
            else:
                message = message.format(reason=", above limit; DEBUG logging is enabled")
                LOGGER.debug(message, extra=extra)

    @staticmethod
    def _get_untranslated_ids(
        ids: Sequence[IdType],
        *,
        mask: Sequence[Optional[str]],
    ) -> List[IdType]:
        seen = set()
        retval = []

        for i, idx in enumerate(ids):
            if idx in seen:
                continue

            if mask[i] is None:
                seen.add(idx)
                retval.append(idx)
                if len(retval) == NUM_SAMPLE_IDS:
                    break

        return retval
